@@ -11,7 +11,7 @@ import random
 
 # Third-party imports
 import psutil
-from libpastis import BrokerAgent, FuzzingEngineDescriptor, EngineConfiguration, BinaryPackage, SASTReport
+from libpastis import BrokerAgent, FuzzingEngineDescriptor, EngineConfiguration, BinaryPackage, SASTReport, ClientAgent
 from libpastis.types import SeedType, FuzzingEngineInfo, LogLevel, Arch, State, SeedInjectLoc, CheckMode, CoverageMode, \
                             ExecMode, AlertData, PathLike, Platform, FuzzMode
 import lief
@@ -47,7 +47,6 @@ class Bcolors:
     UNDERLINE = '\033[4m'
 
 
-
 class PastisBroker(BrokerAgent):
 
     def __init__(self, workspace: PathLike,
@@ -57,7 +56,8 @@ class PastisBroker(BrokerAgent):
                  inject_loc: SeedInjectLoc = SeedInjectLoc.STDIN,
                  sast_report: PathLike = None,
                  p_argv: List[str] = None,
-                 memory_threshold: int = 85):
+                 memory_threshold: int = 85,
+                 start_quorum: int = 0):
         super(PastisBroker, self).__init__()
 
         # Initialize workspace
@@ -115,6 +115,14 @@ class PastisBroker(BrokerAgent):
         self.watchdog = None
         self._threshold = memory_threshold # percent
 
+        # startup quorum
+        self._startup_quorum = start_quorum
+        self._current_quorum = 0
+
+        # Proxy feature
+        self._proxy = None
+        self._proxy_cli = None
+
     def load_engine_addon(self, py_module: str) -> bool:
         desc = load_engine_descriptor(py_module)
         if desc is not None:
@@ -133,15 +141,15 @@ class PastisBroker(BrokerAgent):
     def running(self) -> bool:
         return self._running
 
-    def iter_other_clients(self, client: PastisClient) -> Generator[PastisClient, None, None]:
+    def iter_other_clients(self, cli_id: bytes) -> Generator[PastisClient, None, None]:
         """
         Generator of all clients but the one given in parameter
 
-        :param client: PastisClient client to ignore
+        :param cli_id: netid of the client to ignore
         :return: Generator of PastisClient object
         """
         for c in self.clients.values():
-            if c.netid != client.netid:
+            if c.netid != cli_id:
                 yield c
 
     def new_uid(self) -> int:
@@ -186,7 +194,7 @@ class PastisBroker(BrokerAgent):
         self.statmanager.update_seed_stat(cli, typ)  # Add info only if new
         cli.log(LogLevel.INFO, f"seed {h} [{self._colored_seed_type(typ)}][{self._colored_seed_newness(is_new)}]")
         cli.add_own_seed(seed)  # Add seed in client's seed
-        self.write_seed(typ, cli, seed) # Write seed to file
+        self.write_seed(typ, cli.strid, seed) # Write seed to file
 
         if is_new:
             self._seed_pool[seed] = typ  # Save it in the local pool
@@ -196,10 +204,16 @@ class PastisBroker(BrokerAgent):
 
         # Iterate on all clients and send it to whomever never received it
         if self.broker_mode == BrokingMode.FULL:
-            for c in self.iter_other_clients(cli):
-                if c.is_new_seed(seed):
-                    self.send_seed(c.netid, typ, seed)  # send the seed to the client
-                    c.add_peer_seed(seed)  # Add it in its list of seed
+            self.send_seed_to_all_others(cli.netid, typ, seed)
+
+        if self.is_proxied:  # Forward it to the proxy
+            self._proxy.send_seed(typ, seed)
+
+    def send_seed_to_all_others(self, origin_id: bytes, typ: SeedType, seed: bytes) -> None:
+        for c in self.iter_other_clients(origin_id):
+            if c.is_new_seed(seed):
+                self.send_seed(c.netid, typ, seed)  # send the seed to the client
+                c.add_peer_seed(seed)  # Add it in its list of seed
 
     def add_seed_file(self, file: PathLike, initial: bool = False) -> None:
         p = Path(file)
@@ -212,10 +226,10 @@ class PastisBroker(BrokerAgent):
         if initial:
             self._init_seed_pool[seed] = SeedType.INPUT
 
-    def write_seed(self, typ: SeedType, from_cli: PastisClient, seed: bytes):
+    def write_seed(self, typ: SeedType, cli_id: str, seed: bytes):
         t = time.strftime("%Y-%m-%d_%H:%M:%S", time.localtime())
         elapsed = str(datetime.timedelta(seconds=time.time() - self._start_time)).replace(" day, ", "d:").replace(" days, ", "d:")
-        fname = f"{t}_{elapsed}_{from_cli.strid}_{md5(seed).hexdigest()}.cov"
+        fname = f"{t}_{elapsed}_{cli_id}_{md5(seed).hexdigest()}.cov"
         self.workspace.save_seed(typ, fname, seed)
 
     def hello_received(self, cli_id: bytes, engines: List[FuzzingEngineInfo], arch: Arch, cpus: int, memory: int, hostname: str, platform: Platform):
@@ -229,8 +243,20 @@ class PastisBroker(BrokerAgent):
             if eng.name not in self.engines:
                 self.load_engine_addon(eng.pymodule)
 
-        if self.running: # A client is coming in the middle of a session
-            self.start_client_and_send_corpus(client)
+        if self.running:  # A client is coming in the middle of a session
+            if self._startup_quorum:
+                self._current_quorum += 1
+                if self._current_quorum >= self._startup_quorum:
+                    logging.info(
+                        f"client number quorum ({self._current_quorum}/{self._startup_quorum}) reached, start all of them")
+                    self.start_pending_clients()  # start all pending clients (including the one triggering this func)
+                else:
+                    # Put the client on-hold to wait for quorum
+                    logging.info(f"client {client.strid} on-hold, wait quorum ({self._current_quorum}/{self._startup_quorum})")
+            else:  # there is no quorum so start immediately
+                self.start_client_and_send_corpus(client)
+        else:
+            pass  # Client connection is kept in clients dict for later
 
     def _transmit_pool(self, client, pool) -> None:
         for seed, typ in pool.items():
@@ -340,12 +366,24 @@ class PastisBroker(BrokerAgent):
         # Call the statmanager to wrap-up values
         self.statmanager.post_execution(list(self.clients.values()), self.workspace)
 
+        if self.is_proxied:
+            self._proxy.stop()
+
         if self.sast_report:  # If a SAST report was loaded
             # Write the final CSV in the workspace
             self.sast_report.write_csv(self.workspace.csv_result_file)
             # And also re-write the Klocwork report (that also contains resutls)
             self.sast_report.write(self.workspace.sast_report_file)
 
+    def start_pending_clients(self) -> None:
+        """
+        Start clients that connected but for which we did not yet send a response.
+        """
+        # Send the start message to all clients (already connected)
+        for c in self.clients.values():
+            if not c.is_running():
+                self.start_client_and_send_corpus(c)
+
     def start_client_and_send_corpus(self, client: PastisClient) -> None:
         self.start_client(client)
         # Iterate all the seed pool and send it to the client
@@ -423,7 +461,6 @@ class PastisBroker(BrokerAgent):
                 logging.error(f"In mode {self.ck_mode} but the binary package is not a QBinExport")
                 return
 
-
         # Update internal client info and send him the message
         engine_args_str = engine_args.to_str() if engine_args else ""
         logging.info(f"send start client {client.id}: {package.name} [{engine.NAME}, {covmode.name}, {fuzzmod.name}, {exmode.name}]")
@@ -443,7 +480,6 @@ class PastisBroker(BrokerAgent):
                         self.inject,
                         self.sast_report.to_json() if self.sast_report else b"")
 
-
     def _find_configuration(self, engine: FuzzingEngineDescriptor) -> Optional[EngineConfiguration]:
         """
         Find a configuration for the engine. It will iterate all configuration
@@ -499,11 +535,13 @@ class PastisBroker(BrokerAgent):
         self.workspace.status = WorkspaceStatus.RUNNING
         logging.info("start broking")
 
-        # Send the start message to all clients
-        if self._running:  # If we want to run now (cmdline mode)
-            for c in self.clients.values():
-                if not c.is_running():
-                    self.start_client(c)
+        if self.is_proxied and self._proxy_cli:
+            self._running = False  # disable running wait start broker
+            self._proxy.send_hello([self._proxy_cli])
+        else:
+            # Send the start message to all clients (already connected)
+            if self._running:  # If we want to run now (cmdline mode)
+                self.start_pending_clients()
 
     def run(self, timeout: int = None):
         self.start()
@@ -633,3 +671,54 @@ class PastisBroker(BrokerAgent):
             logging.warning(f"Threshold reached: {mem.percent}%")
             return False
         return True
+
+    @property
+    def is_proxied(self) -> bool:
+        """
+        Get if the broker as a proxy to another broker.
+        """
+        return self._proxy is not None
+
+    def set_proxy(self, ip: str, port: int, py_module: str) -> bool:
+        self._proxy = ClientAgent()
+
+        # Load the engine info to impersonate
+        desc = load_engine_descriptor(py_module)
+        if desc is None:
+            logging.error("Cannot load FuzzingEngineDescriptor")
+            return False
+        else:
+            self._proxy_cli = FuzzingEngineInfo(desc.NAME, desc.VERSION, py_module)
+
+        self._proxy.register_start_callback(self._proxy_start_received)
+        self._proxy.register_seed_callback(self._proxy_seed_received)
+        self._proxy.register_stop_callback(self._proxy_stop_received)
+
+        logging.info(f"connect to broker {ip}:{port}")
+        self._proxy.connect(ip, port)
+        self._proxy.start()
+
+    def _proxy_start_received(self, fname: str, binary: bytes, engine: FuzzingEngineInfo, exmode: ExecMode,
+                              fuzzmode: FuzzMode, chkmode: CheckMode, covmode: CoverageMode, seed_inj: SeedInjectLoc,
+                              engine_args: str, argv: List[str], sast_report: str = None):
+        # FIXME: Use parameters received
+        logging.info("[PROXY] start received !")
+        self._running = True
+        if self._running:
+            self.start_pending_clients()
+
+    def _proxy_seed_received(self, typ: SeedType, seed: bytes):
+        # Forward the seed to underlying clients
+        logging.info(f"[PROXY] seed {typ.name} received forward to agents")
+
+        # Save the seed locally
+        self.write_seed(typ, "PROXY", seed)
+        self._seed_pool[seed] = typ  # add it to the pool
+        self._init_seed_pool[seed] = typ  # also consider it as initial corpus
+
+        # Forward it to all clients
+        self.send_seed_to_all_others(b"PROXY", typ, seed)
+
+    def _proxy_stop_received(self):
+        logging.info(f"[PROXY] stop received!")
+        self._stop = True
